/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Forecastfox.
 *
 * The Initial Developer of the Original Code is
 * Jon Stritar <jstritar@MIT.EDU>.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Jon Stritar <jstritar@MIT.EDU>
 * Richard Klein <richwklein@mchsi.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const CLASS_ID = Components.ID("{6E33307F-98BF-42da-98CA-DCDED4282E84}");
const CLASS_NAME = "Forecastfox Manager Component";
const CONTRACT_ID = "@ensolis.com/forecastfox/manager;1";
const ffIError = Components.interfaces.ffIError;
const ffIDisk = Components.interfaces.ffIDisk;
const nsIRequest = Components.interfaces.nsIRequest;

/******************************************************************************
 * ffManager component
 ******************************************************************************/
function ffManager() {};
ffManager.prototype = {
  _partner: "forecastfox",
  _url: "http://forecastfox.accuweather.com/adcbin/forecastfox/weather_data.asp?",
  _firstTime: true,
  _parser: null,
  _profiles: null,
  _disk: null,      
  _ping: null,
  _guid: null,
  _branch: null,  
  _obs: null,
  _timer: null,
  _bundle: null,
  _ps: null,     
  _groups: null,
  _requests: null,
  _persist: null,
  _radarFile: null,
  _firstError: true,

   
  //readonly attributes fo UI
  get partner() { return this._partner; },
  get url() { return this._url; },
  get firstTime() { return this._firstTime; },
  
  get parser() { 
    if (!this._parser) {
      this._parser = Components.classes["@ensolis.com/forecastfox/parser;1"].getService(Components.interfaces.ffIParser);
      try {
        this._parser.start();
      } catch (e) {
        this.disk.recordError("Message: Error starting the parser module.", e);
        this._parser = null;
      };
    };
    return this._parser;
  },

  get profiles() {
    if (!this._profiles) {
      this._profiles = Components.classes["@ensolis.com/forecastfox/profiles;1"].getService(Components.interfaces.ffIProfiles);
      try {
        var error = this._profiles.start();
        if (error.severity != ffIError.SEVERITY_INFO)
          this._ps.alert(this.getWindow(), error.description, error.tooltip);
      } catch (e) {
        this.disk.recordError("Message: Error starting the profiles module.", e);
        this._profiles = null;
      };
    };
    return this._profiles;
  },
  
  get disk() {
    if (!this._disk) {
      this._disk = Components.classes["@ensolis.com/forecastfox/disk;1"].getService(Components.interfaces.ffIDisk);
      try {
        this._disk.start();
      } catch(e) {
        //XXX need to localize
        this._ps.alert(this.getWindow(), "Disk Error", "An error occured starting the Disk I/0 Component.");
        this._disk = null;
      };
    };  
    return this._disk; 
  },
  
  get ping() {
    if (!this._ping) {
      this._ping = Components.classes["@ensolis.com/forecastfox/ping;1"].getService(Components.interfaces.ffIPing);
      try {
        this._ping.start();
      } catch (e) {
        this.disk.recordError("Message: Error starting the ping module.", e);
        this._ping = null;
      };
    };
    return this._ping;  
  },
  
  get guid() {
    if (!this._guid) {
      try {
        var app = Components.classes["@mozilla.org/xre/app-info;1"].getService(Components.interfaces.nsIXULAppInfo);
        this._guid = app.ID;
      } catch(e) {
        var pbs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
        var appBranch = pbs.getBranch(null);
        try {
          this._guid = appBranch.getCharPref("app.id");
        } catch(e) {
          this._guid = "{86c18b42-e466-45a9-ae7a-9b95ba6f5640}";
        };
      };
      // get rid of extra characters that are in some people's app.id pref
      this._guid = this._guid.match(/{.*}/)[0];
    };
    return this._guid;
  },
  
  get branch() {
    if (!this._branch) {
      var pbs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
      this._branch = pbs.getBranch("forecastfox."); 
    };
    return this._branch;
  },
    
  start: function()
  {
    //setup components
    this._obs = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
    this._timer = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);
    var sbs = Components.classes["@mozilla.org/intl/stringbundle;1"].getService(Components.interfaces.nsIStringBundleService);    
    this._bundle = sbs.createBundle("chrome://forecastfox/locale/forecastfox.properties");
    this._ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
      
    //ping server
    try {
      this.ping.open();
    } catch (e) {
      this.disk.recordError("Message: Error pinging the server.", e);
    };

    //load profiles
    try {
      this._profiles = Components.classes["@ensolis.com/forecastfox/profiles;1"].getService(Components.interfaces.ffIProfiles);
      var error = this._profiles.start();
      if (error.severity != ffIError.SEVERITY_INFO)
        this._ps.alert(this.getWindow(), error.description, error.tooltip);
    } catch (e) {
      this.disk.recordError("Message: Error starting the profiles module.", e);
      this._profiles = null;
    }
    
    //setup current group
    //XXX create a new group if we split updates
    this._groups = new Object();
    this._groups["current"] = new Object();
    this._groups["current"].targets = ["swa", "radar", "current", "days", "nights"];
    this._groups["current"].frequency = "general.freq";
    this._groups["current"].last = "general.last";
    this._groups["current"].cache = "general.cache";
    this._groups["current"].from = "server";
    this._groups["current"].errorSeverity = ffIError.SEVERITY_INFO;
    this._groups["current"].errorLabel = "";
    this._groups["current"].errorTooltip = "";
    this._groups["current"].loading = false;
      
    //add shutdown observer
    this._obs.addObserver(this, "xpcom-shutdown", false);
    this._obs.addObserver(this, "forecastfox:profile", false);

    //add prefernce observer
    var pbi = this.branch.QueryInterface(Components.interfaces.nsIPrefBranchInternal);
    pbi.addObserver("", this, false);
  },
  
  stop: function()
  {
    //cancel request
    this.cancel();
     
    //stop timers
    this._timer.cancel();
    for (var j=0; j<this._requests.length; j++)
      this._requests[j].timeout.cancel();
                  
    //clean up held observers etc to avoid leaks.
    try {
      this._obs.removeObserver(this, "xpcom-shutdown");
      this._obs.removeObserver(this, "forecastfox:profile");
      var pbi = this._branch.QueryInterface(Components.interfaces.nsIPrefBranchInternal);
      pbi.removeObserver("", this);
      pbi = null;
    } catch(e) {};
              
    //stop components
    this._ping.stop();
    this._parser.stop();
    this._profiles.stop();
    this._disk.stop();
        
    //destroy variables
    this._partner = null;
    this._url = null;
    this._firstTime = null;
    this._parser = null;
    this._profiles = null;
    this._disk = null;    
    this._ping = null;
    this._guid = null;
    this._branch = null;
    this._obs = null;
    this._timer = null;
    this._bundle = null;
    this._ps = null;
    this._groups = null;
    this._requests = null;
    this._persist = null;
    this._radarFile = null;
    this._firstError = null;
  },
    
  openLink: function(aURL, aWhere)
  {
    //XXX may need to change this code if main window of a supported app is not a browser
    var win = this.getWindow();
    var browser = win.document.getElementById("content");
    switch (aWhere) {
      case "window":
        switch (this.guid) {
          case "{ec8030f7-c20a-464f-9b0e-13a3a9e97384}":
          case "{3db10fab-e461-4c80-8b97-957ad5f8ea47}":
          default:        
            win.openDialog("chrome://browser/content/browser.xul", "_blank", "chrome,all,dialog=no", aURL, null, null);
            break;
          case "{86c18b42-e466-45a9-ae7a-9b95ba6f5640}":
            win.openDialog("chrome://navigator/content/navigator.xul", "_blank", "chrome,all,dialog=no", aURL, null, null);
            break;
        };      
        break;
      case "tab":
      case "tabshifted":
        var tab = browser.addTab(aURL);
        if (aWhere == "tab") {
          browser.selectedTab = tab;
          win.content.focus();
        };        
        break;
      case "current":
      default: 
        browser.loadURI(aURL);
        win.content.focus();
        break;
    };
  },
  
  getWindow: function()
  {
    //XXX may need to change this code if main window of a supported app is not "navigator:browser"
    var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator);
    var ww = Components.classes["@mozilla.org/embedcomp/window-watcher;1"].getService(Components.interfaces.nsIWindowWatcher);
    var top = wm.getMostRecentWindow("navigator:browser");
    if (!top) {
      switch (this.guid) {
        case "{ec8030f7-c20a-464f-9b0e-13a3a9e97384}":
        case "{3db10fab-e461-4c80-8b97-957ad5f8ea47}":
        default:        
          top = ww.openWindow(null, "chrome://browser/content/browser.xul", "_blank", "chrome,all,dialog=no", "about:blank");
          break;
        case "{86c18b42-e466-45a9-ae7a-9b95ba6f5640}":
          top = ww.openWindow(null, "chrome://navigator/content/navigator.xul", "_blank", "chrome,all,dialog=no", "about:blank");
          break;
      };
    };
    return top;  
  },
    
  load: function(aForced)
  {
    //initialize
    if (this._firstTime) {
      // XXX this._firstTime should be set to true after starting... otherwise, during profiles.start
      //     methods in here can be called when things aren't initialized.
      //set not first time
      this._firstTime = false;
      this.start();
    };
          
    //reset last prefs if forced
    if (aForced) {
      for (i in this._groups) {
        if (this.branch.prefHasUserValue(this._groups[i].last))
          this.branch.clearUserPref(this._groups[i].last);
      };
    };
          
    //get data
    this._getData();
  },

  getError: function(aGroup)
  {
    var error = Components.classes["@ensolis.com/forecastfox/error;1"].createInstance(Components.interfaces.ffIError);
    error.init(this._groups[aGroup].errorSeverity, "@ensolis.com/forecastfox/manager;1", this._groups[aGroup].errorLabel, this._groups[aGroup].errorTooltip);      
    return error;
  },
  
  cancel: function()
  {
    //if not loaded yet do nothing
    if (this._firstTime)
      return;
    
    //stop timer
    this._timer.cancel();
        
    //cancel pending persists
    if (this._persist) {
      if (this._persist.currentState < 3)
        this._persist.cancelSave(); 
    };
      
    //loop through groups
    if (this._requests) {
      for (i in this._groups) {
        //match request with group
        for (var j=0; j<this._requests.length; j++) {
          if (this._groups[i].from != this._requests[j].from)   
            continue;
        
          //cancel any pending requests
          if (this._requests[j].readyState < 4) {
            this._requests[j].abort();
            this._requests[j].timeout.cancel();
            this._obs.notifyObservers(this, "forecastfox:manager", "end");
            break;
          };
        };
      };
    };
  },
    
  _getData: function()
  {
    //do nothing if location not set
    if (!this._checkLocID())
      return;
      
    //notify observers of update 
    //XXX move back into the for loop when we seperate current and forecast
    this._obs.notifyObservers(this, "forecastfox:manager", "start");
    
    for (i in this._groups) {
      
      //check if we need to come from server
      this._checkServer(i);

      //clear last error
      this._groups[i].errorSeverity = ffIDisk.SEVERITY_INFO;
      this._groups[i].errorLabel = "";
      this._groups[i].errorTooltip = "";
    
      //mark for loading
      this._groups[i].loading = true;
    };
    
    //build out our data requests
    this._requestData();
  },
  
  _requestData: function()
  {
    var cache = new Object();
    var server = false;
    var j, request, query;
    var component = this;

    //cancel any requests
    this.cancel();
        
    //build new requests
    this._requests = new Array();    
    for (i in this._groups) {
      if (this._groups[i].from == "server") {
        if (!server) {
          query = this._getServerURL();
          request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
          request.timeout = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);
          request.overrideMimeType("text/xml");
          request.from = "server";
          request.error = false;
          request.onerror = function() { component._receiveData(this); };            
          request.onload = function() { component._receiveData(this); };            
          request.open("GET", query, true);   
          request.channel.loadFlags |= nsIRequest.LOAD_BYPASS_CACHE;                   
          this._requests.push(request);
          server = true;                 
        };
      } else {
        if (!cache[this._groups[i].from]) {
          query = this._getFileURL(this._groups[i].from);
          request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
          request.timeout = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);
          request.overrideMimeType("text/xml");
          request.from = this._groups[i].from;
          request.error = false;          
          request.onerror = function() { component._receiveData(this); };            
          request.onload = function() { component._receiveData(this); };            
          request.open("GET", query, true);            
          this._requests.push(request);
          cache[this._groups[i].from] = true;              
        };
      };
    };
    
    //send requests
    for (j=0; j<this._requests.length; j++) {
      this._requests[j].timeout.initWithCallback(this, 20 * 1000, Components.interfaces.nsITimer.TYPE_ONE_SHOT);
      this._requests[j].send(null);
    };
  },
  
  _receiveData: function(aXMLReq)
  {
    var parseError = false;
    var hasError = false;
    
    //cancel timeout
    aXMLReq.timeout.cancel();
    
    for (i in this._groups) {
    
      //skip if not the same from
      if (this._groups[i].from != aXMLReq.from)
        continue;
      
      //mark loading complete
      this._groups[i].loading = false;
      
      //check if we actual received anything
      if (aXMLReq.responseText == "") {
        this._setError(i, ffIError.SEVERITY_ERROR, "ff.error.doc", null, null);
        hasError = true;          
        continue; 
      };
                 
      //check status if from server
      try {
        var status = aXMLReq.status;
      } catch (e) {
        // exception is thrown when the server can't be contacted
        this._setError(i, ffIError.SEVERITY_ERROR, "ff.error.doc", null, null);
        hasError = true;
        status = -1;
      };
      
      if (aXMLReq.from == "server") {
        if (status == 407) {
          //XXX do authentication
          this._setError(i, ffIError.SEVERITY_ERROR, "ff.error.auth", null, null);
          hasError = true;
          continue;
        };  
        if (status > 200) {
          this._setError(i, ffIError.SEVERITY_ERROR, "ff.error.net", null, null);
          hasError = true;
          continue;
        };
      };
      
      //parse document
      var targets = this._groups[i].targets;
      for (var j=0; j<targets.length; j++) {
        try {
          var error = this.parser.open(aXMLReq.responseXML, targets[j], (j==0));
          
          //check for error
          if (error.severity != ffIError.SEVERITY_INFO) {
            this._setError(i, error.severity, null, error.label, error.tooltip);
            hasError = true;
            parseError = true;      
            break;
          };
        } catch (e) {
          this.disk.recordError("Message: It looks like parser.open returned a null ffIError.", e);
        };
      };
      
      //go to next group
      if (parseError)
        continue;
      
      //do notifications when from server
      if (aXMLReq.from == "server") {
          //save radar image
          this._radar();
            
          //do sliders
          this._slider();
      };
    };
    
    //notify update complete
    //XXX make group specific and place back in loop when seperating current and forecast
    if (!hasError && !parseError) {
      var source = this.parser.getValue("radar", "image", null);
      var enabled = this.branch.getBoolPref("radar.panel.enabled");
      if (aXMLReq.from != "server" || source == "" || !enabled)
        this._obs.notifyObservers(this, "forecastfox:manager", "end");
    };
    
    //set error
    aXMLReq.error = hasError;

    //force from server next time    
    if (hasError) {   
      if (aXMLReq.from != "server") {
        for (i in this._groups) {
          if (this._groups[i].from == aXMLReq.from && 
              this.branch.prefHasUserValue(this._groups[i].last))
            this.branch.clearUserPref(this._groups[i].last);            
        };
      };      
    } else {
      if (aXMLReq.from == "server") {
        var freq;   
        //get expiration
        try {
          var expires = aXMLReq.getResponseHeader("expires");
          var next = new Date(expires);
          var current = new Date();
          var random = 1 + Math.random()*2;
          freq = Math.floor((next-current)/(60*1000) + random);
          if (-5 < freq && freq <= 0)
            freq = 5
          if (freq < 0 || freq > 30)
            freq = 30;
        } catch(e) {
          freq = 30;
        };
        
        //save data 
        this._saveData(aXMLReq.responseXML, freq);  
      };    
    };
    
    //schedule
    switch (this._checkSchedule()) {
      case "none":
        break;
      case "error":
        // if this is the first error, refresh after 10 seconds.
        // if this is not the first, refresh after 2 minutes.
        if (this._firstError) {
          this._firstError = false;
          this._schedule(10*1000);
        } else
          this._schedule(2*60*1000);
        break;
      case "normal":
      default:
        this._firstError = true;
        this._schedule();
        break;
    };
  },
  
  _radar: function()
  {  
    var ios = Components.classes["@mozilla.org/network/io-service;1"].createInstance(Components.interfaces.nsIIOService);
    var name = "radarcache-" + this.profiles.current + ".gif";
    var source = this.parser.getValue("radar", "image", null);
    var enabled = this.branch.getBoolPref("radar.panel.enabled");
     
 	  if (source != "" && enabled) {
      //convert url to uri and get file
      var uri = ios.newURI(source, null, null);
      this._radarFile = this.disk.get(name, ffIDisk.TYPE_TEMP);

      var flags = 0x02 | 0x32;
      this._persist = Components.classes["@mozilla.org/embedding/browser/nsWebBrowserPersist;1"].createInstance(Components.interfaces.nsIWebBrowserPersist); 				      
      this._persist.persistFlags |= flags;
      this._persist.progressListener = this;
      this._persist.saveURI(uri, null, null, null, null, this._radarFile);
      
      //set cache
      this.branch.setCharPref("radar.cache", name);
    } else {
      this.branch.setCharPref("radar.cache", "");
      // XXX re-add end notification if we seperate current and forecast again
    }; 
  },
  
  _slider: function()
  {
    var type = null;
    var freq = 1;
    var count = 1;
    var enabled = false;
    var active = false;
    var icon = "";
    var title = "";
    var text = "";
    var url = "";
    
    if (this.parser.getValue("swa", "active", null) == "true")
      active = true;
    else
      active = false;
 
    //determine if we are showing swa
    if (active) {
      //always increment count if active
      count = this.branch.getIntPref("swa.slider.count");
      this.branch.setIntPref("swa.slider.count", count + 1);
      
      //calculate if its time to check
      enabled = this.branch.getBoolPref("swa.slider.enabled");
      freq = this.branch.getIntPref("swa.slider.freq");
      
      if (enabled && count == 1)
        type = "swa";
      
      if (count >= freq)
        this.branch.setIntPref("swa.slider.count", 1);
    } else
      this.branch.setIntPref("swa.slider.count", 1);
    
    //always increment count for cc
    count = this.branch.getIntPref("cc.slider.count");
    this.branch.setIntPref("cc.slider.count", count + 1);
    
    //check if we are doing current
    if (!type) {
      enabled = this.branch.getBoolPref("cc.slider.enabled");
      freq = this.branch.getIntPref("cc.slider.freq");     
      if (enabled && count >= freq)
        type = "current";
    };

    //do neither
    if (!type)
      return;
         
    //set values based on type
    if (type == "current") {
      this.branch.setIntPref("cc.slider.count", 1);    
      url = this.parser.getValue("current", "url", null);
      switch (this.branch.getIntPref("cc.slider.display")) {
        case 0:
          icon = "chrome://forecastfox/skin/images/large/" + this.parser.getValue("current", "icon", null) + ".png";  
          break;
        case 1:
          title = this.parser.setLabel(this.branch.getComplexValue("cc.slider.title",Components.interfaces.nsIPrefLocalizedString).data, "current", null);
          text = this.parser.setLabel(this.branch.getComplexValue("cc.slider.label",Components.interfaces.nsIPrefLocalizedString).data, "current", null);
          break;
        case 2:
          icon = "chrome://forecastfox/skin/images/large/" + this.parser.getValue("current", "icon", null) + ".png";  
          title = this.parser.setLabel(this.branch.getComplexValue("cc.slider.title",Components.interfaces.nsIPrefLocalizedString).data, "current", null);
          text = this.parser.setLabel(this.branch.getComplexValue("cc.slider.label",Components.interfaces.nsIPrefLocalizedString).data, "current", null);
          break;
      };
    } else {
      url = this.parser.getValue("swa", "url", null);
      switch (this.branch.getIntPref("swa.slider.display")) {
        case 0:
          icon = "chrome://forecastfox/skin/images/large/" + this.parser.getValue("swa", "icon", null) + ".png";
          break;
        case 1:
          title = this.parser.setLabel(this.branch.getComplexValue("swa.slider.title",Components.interfaces.nsIPrefLocalizedString).data, "swa", null);
          text = this.parser.setLabel(this.branch.getComplexValue("swa.slider.label",Components.interfaces.nsIPrefLocalizedString).data, "swa", null);       
          break;
        case 2:
          icon = "chrome://forecastfox/skin/images/large/" + this.parser.getValue("swa", "icon", null) + ".png";
          title = this.parser.setLabel(this.branch.getComplexValue("swa.slider.title",Components.interfaces.nsIPrefLocalizedString).data, "swa", null);
          text = this.parser.setLabel(this.branch.getComplexValue("swa.slider.label",Components.interfaces.nsIPrefLocalizedString).data, "swa", null);       
          break;
      };
    };

    //do alert  
    try {
      var alerts = Components.classes["@mozilla.org/alerts-service;1"].getService(Components.interfaces.nsIAlertsService);
      alerts.showAlertNotification(icon, title, text, true, url, this);
    } catch(e) {};         
  },
    
  _saveData: function(aDoc, aFreq)
  {
    var freq = 0;
    var profile = this.profiles.current;
    var current = (new Date()).valueOf();
    var name;
    var gFreq;
    
    //figure out file name
    for (i in this._groups) {
      if (this._groups[i].from != "server")
        continue;
      
      gFreq = this.branch.getIntPref(this._groups[i].frequency);
      if (gFreq < 0) {
        gFreq = 30;
        this.branch.setIntPref(this._groups[i].frequency, gFreq);
      };
      if (gFreq >= freq) {
        freq = gFreq
        name = i + "cache-" + profile + ".xml"; 
      };
    };
    
    //flush data to disk 
    var file = this.disk.get(name, ffIDisk.TYPE_CACHE);
    this.disk.write(aDoc, file, false);
    
    //update the preferences    
    for (i in this._groups) {
      if (this._groups[i].from != "server")
        continue;
      this.branch.setCharPref(this._groups[i].cache, name);
      this.branch.setCharPref(this._groups[i].last, current);
      this.branch.setIntPref(this._groups[i].frequency, aFreq);
    };    
  },
    
  _setError: function(aGroup, aSeverity, aName, aDescription, aTooltip)
  {
    this._groups[aGroup].errorSeverity = aSeverity;
    this._groups[aGroup].errorLabel = (aName) ? this._bundle.GetStringFromName(aName + ".label") : aDescription;
    this._groups[aGroup].errorTooltip = (aName) ? this._bundle.GetStringFromName(aName + ".tooltip") : aTooltip;
    
    // XXX now that both groups update together, there are always two 
    //     "error" notifications whenever there is an error
    this._obs.notifyObservers(this, "forecastfox:manager", "error");
  },
  
  _checkSchedule: function()
  {    
    //first check all requests are complete
    for (var i=0; i<this._requests.length; i++) {
      if (this._requests[i].readyState < 4)
        return "none";
    };
    
    //now check if we had any errors
    for (i=0; i<this._requests.length; i++) {
      if (this._requests[i].error)
        return  "error";
    };

    return "normal";
  },
  
  _checkLocID: function()
  {
    var isSet = this.branch.prefHasUserValue("general.locid");
    var locid = this.branch.getCharPref("general.locid");
      
    //if the locid hasn't been set yet, prompt for
    if (!isSet || locid == "00000" || locid == "") {
      var wm = Components.classes['@mozilla.org/appshell/window-mediator;1'].getService(Components.interfaces.nsIWindowMediator);
      var options = wm.getMostRecentWindow("forecastfox:options");
      if (options)
        options.focus();
      else    
        this.getWindow().openDialog("chrome://forecastfox/content/options.xul", "options", "chrome,modal");    
      return false; 
    };
    
    return true;
  },

  _checkServer: function(aName)
  {
    //current date/time
    var current = (new Date()).valueOf();
    var freq = this.branch.getIntPref(this._groups[aName].frequency);
    
    //if frequency is way out of wack set to 30
    if (freq < 0)
      freq = 30
      
    //then get the time for the next refresh.
    var last = Number(this.branch.getCharPref(this._groups[aName].last));
    var next = last + (freq * 60 * 1000);

	  //refresh if it has been the allowed time, or if the last refresh occured in the future
    if ((last > current) || ((next - current) <= 100))
      this._groups[aName].from = "server";
    else if (this._checkCache(aName))
      this._groups[aName].from = this.branch.getCharPref(this._groups[aName].cache);
    else
      this._groups[aName].from = "server";
  },
  
  _checkCache: function(aName)
  {
    var pref = this.branch.getCharPref(this._groups[aName].cache);
    if (!pref)
      return false;
      
    if (!this._getFileURL(pref))
      return false;
    
    return true;
  },
      
  _getFileURL: function(aName)
  {
    var file = this.disk.get(aName, ffIDisk.TYPE_CACHE);
    if (!file.exists())
      return null;
      
    //convert to a URI and return spec
    var ios = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
    return ios.newFileURI(file).spec;  
  },
  
  _getServerURL: function()
  {
    var query = this.url + "location=" + this.branch.getCharPref("general.locid");

    //set units
    var units = this.branch.getIntPref("general.units");   
    if (units == 1)
      query += "&metric=0";
    else if (units == 2)
      query += "&metric=3";
    else
      query += "&metric=1";
    
    //add partner  
    query += "&partner=" + this.partner;
    
    //return string
    return query
  },
  
  _schedule: function(aForce)
  {
    var actual = null;
    
    if (aForce)
      actual = aForce;
    else {
      var current = (new Date()).valueOf();
      var next = new Array();
      var span = this.branch.getIntPref("schedule.span") * 60 * 1000;
      var bump = this.branch.getIntPref("schedule.bump") /100;
      var freq, value;
          
      //get time till next refresh for each category.
      for (i in this._groups) {
        if (this._groups[i].loading)
          return;
        freq = this.branch.getIntPref(this._groups[i].frequency);
        value = Number(this.branch.getCharPref(this._groups[i].last));
        value += (freq * 60 * 1000);
        value -= current;
        next.push(value)
      }
      
      //sort by time
      next.sort(function (a, b) { return (a - b) });

      //if next is within span * bump go to next
      actual = next[0];
      for (var j=1; j<next.length; j++) {
        if (next[j] - actual < span) {
          actual = next[j];
          span *= bump;
        } else
          break;
      };
      
      //make sure we have a positive time
      if (actual < 0)
        actual = 100;

      actual += 100;
      
    };

    //cancel old timers
    this._timer.cancel();

    // Then schedule the timer
    this._timer.initWithCallback(this, actual, Components.interfaces.nsITimer.TYPE_ONE_SHOT);
  },

  ///////////////////////////
  // nsITimerCallback
  notify: function(timer)
  {
    if (timer == this._timer)
      this._getData();
    else {
      var index = -1;

      for (var j=0; j<this._requests.length; j++) {
        if (this._requests[j].timeout != timer)
          continue;
        
        index = j;
        
        //abort the request    
        this._requests[j].abort();
        
        //send error and set to not loading
        for (i in this._groups) {
          if (this._groups[i].from == this._requests[j].from) {
            this._setError(i, ffIError.SEVERITY_ERROR, "ff.error.timeout", null, null);
            this._groups[i].loading = false;
          };
        };
        break;
      };
      
      if (index != -1)
        this._requests.splice(index, 1);
      
      //schedule update
      this._schedule(2*60*1000);
    };
  },
    
  ///////////////////////////
  // nsIWebProgressListener
  onLocationChange: function(webProgress, request, location) {},
  onProgressChange: function(webProgress, request, curSelfProgress, maxSelfProgress, curTotalProgress, maxTotalProgress ) {},
  onSecurityChange: function(webProgress, request, state) {},
  onStatusChange: function(webProgress, request, status, message) {},
  onStateChange: function(webProgress, request, stateFlags, status)
  {
    var stateDone = Components.interfaces.nsIWebProgressListener.STATE_STOP;
    if (stateDone == (stateFlags & stateDone)) {
      //move file to correct location
      var name = "radarcache-" + this.profiles.current + ".gif";    
      var file = this.disk.get(name, ffIDisk.TYPE_CACHE);
      this.disk.move(this._radarFile, file);
      
      //remove old entry from cache
      try {
        var ios = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
        var cache = Components.classes["@mozilla.org/image/cache;1"].getService(Components.interfaces.imgICache);
        var url = ios.newFileURI(file);
        cache.removeEntry(url);
      } catch(e) {};
      
      //notify frontend
      this._obs.notifyObservers(this, "forecastfox:manager", "end");
    }
  },


  ///////////////////////////
  // nsIAlertListener
  onAlertFinished: function() {},
  onAlertClickCallback: function(aCookie)
  {
    var where = this.branch.getCharPref("links.alert");
    this.openLink(aCookie, where);
  },

  /////////////////////////////
  // nsIObserver
  observe: function(aSubject, aTopic, aData)
  {
    var enabled, prefix;
    
    //do nothing until the first load
    if (this._firstTime)
      return;
      
    switch (aTopic) {
      case "forecastfox:profile":
        //profile loaded so send notification and schedule update
        this._obs.notifyObservers(this, "forecastfox:manager", "profile");
        this._schedule(200);
        break;        
      case "nsPref:changed":
        if (this.profiles.batch)
          return;              
        switch (aData) {      
          case "profile.current":
            this.profiles.load(this.branch.getCharPref(aData));
            break;
          case "profile.switch.enabled":
            enabled = this.branch.getBoolPref(aData);
            if (enabled)
              this.profiles.rotate(this.profiles.ROTATE_START);
            else
              this.profiles.rotate(this.profiles.ROTATE_STOP);
            break;
          case "profile.switch.delay":
            enabled = this.branch.getBoolPref("profile.switch.enabled");
            if (enabled)
              this.profiles.rotate(this.profiles.ROTATE_START);
            break;
          case "general.locid":
          case "general.units":
            for (i in this._groups) {
              if (this.branch.prefHasUserValue(this._groups[i].last))
                this.branch.clearUserPref(this._groups[i].last);
            };
            this._schedule(200);
            break;
          case "general.bar":
          case "general.position":   
            this._obs.notifyObservers(this, "forecastfox:manager", "move");
            break;
          case "swa.panel.enabled":
          case "cc.panel.enabled":
          case "dayt.panel.enabled":
          case "dayf.panel.enabled":             
            this._obs.notifyObservers(this, "forecastfox:manager", "enable");
            break;
          case "radar.panel.enabled":
            if (this.branch.prefHasUserValue(this._groups["current"].last))
              this.branch.clearUserPref(this._groups["current"].last);
            this._schedule(200);
            break;
          case "radar.panel.display":
          case "radar.panel.label":
          case "radar.tooltip.display":
          case "radar.tooltip.label":
          case "radar.tooltip.enabled":
          case "swa.panel.display":
          case "swa.panel.label":
          case "swa.tooltip.enabled":
          case "swa.tooltip.display":
          case "swa.tooltip.label":                    
          case "cc.panel.display":
          case "cc.panel.label":
          case "cc.tooltip.enabled":
          case "cc.tooltip.display":
          case "cc.tooltip.label":
          case "dayt.panel.display":
          case "dayt.tooltip.enabled":
          case "dayt.tooltip.display":
          case "dayt.tooltip.label":
          case "dayt.panel.mode":
          case "dayt.panel.switch":
          case "dayt.panel.label":
          case "dayf.panel.display":
          case "dayf.panel.label":
          case "dayf.tooltip.enabled":
          case "dayf.tooltip.display":
          case "dayf.tooltip.label":
          case "dayf.panel.mode":
          case "dayf.panel.days":
            prefix = aData.substring(0, aData.indexOf("."));
            this._obs.notifyObservers(this, "forecastfox:manager", "data," + prefix);
            break;                       
        };            
        break;
      case "xpcom-shutdown":
		this.stop();
        break;
    }
  },

  QueryInterface: function (iid)
  {
    if (!iid.equals(Components.interfaces.ffIManager) &&
          !iid.equals(Components.interfaces.nsIAlertListener)
          && !iid.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
}

/******************************************************************************
 * XPCOM Functions for construction and registration
 ******************************************************************************/
var gModule = {
  _firstTime: true,
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  {
    if (this._firstTime) {
      this._firstTime = false;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    };
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME, CONTRACT_ID, aFileSpec, aLocation, aType);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType)
  {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);        
  },
  
  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return gFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) { return true; }
};

var gFactory = {
  createInstance: function (aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    return (new ffManager()).QueryInterface(aIID);
  }
};

function NSGetModule(aCompMgr, aFileSpec) { return gModule; }
